from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np


#UI_base.Add_element(InteractiveScatter(...), sidebar, stretch)
class InteractiveScatter(pg.GraphicsLayoutWidget):#canvas object

    def __init__(self,default_x='index', default_y='score', coloration_param='list', *args, **kwargs):#
        super().__init__(*args, **kwargs)

        self.default_x = default_x
        self.default_y = default_y
        self.coloration_param = coloration_param
        self.global_coloration = True

        self.storage_manager_groups = []

        self.initialize_plot()

        self.scatter_clicked_event = None
        self.scatter_double_clicked_event = None


    def scatter_clicked(self, plot, points):
        print('clicked')
        if len(points) > 0:
            self.clicked_generation = points[-1]._data[0]
            self.clicked_score = points[-1]._data[1]

            self.clicked_sm = points[-1]._data[3]
            self.clicked_smg = points[-1]._data[4]
            print(self.clicked_sm)

            self.scatter2.setData(x=[self.clicked_generation], y=[self.clicked_score])  # set second scatter to new selection
            self.scatter2.data[0][3] = self.clicked_sm
            if self.scatter_clicked_event is not None:
                self.scatter_clicked_event(self.clicked_sm)

    def scatter_double_clicked(self, plot, points):#scatter2

        print('double clicked', self.clicked_sm)

        #evo_id = tab.data[-3, tab.clicked_sm][0]
        # print(evo_id)

        sm = self.clicked_sm#self.clicked_smg['id==' + str(int(self.clicked_sm))][0]

        txt = open(sm.absolute_path + sm.config_file_name, 'r').read()

        layout = QVBoxLayout()
        pte = QPlainTextEdit()
        pte.setPlainText(txt)
        pte.setReadOnly(True)
        layout.addWidget(pte)

        dlg = QDialog()
        #dlg.setWindowTitle(sm.absolute_path + sm.config_file_name)
        dlg.setLayout(layout)
        dlg.resize(1200, 800)
        dlg.exec()

        if self.scatter_double_clicked_event is not None:
            self.scatter_double_clicked_event(self.clicked_sm)

    def change_axis_param(self, axis_name, param):

        dx = self.default_x
        dy = self.default_y
        dc = self.coloration_param

        try:
            if axis_name == 'bottom':
                self.default_x = param
                self.plot.getAxis(axis_name).setLabel(text=param)

            if axis_name == 'left':
                self.default_y = param
                self.plot.getAxis(axis_name).setLabel(text=param)

            if axis_name == 'color':
                self.coloration_param = param

            self.refresh_data()

            return param
        except:
            self.default_x = dx
            self.default_y = dy
            self.coloration_param = dc



    def get_all_params(self):
        result_dict = {}
        for smg in self.storage_manager_groups:
            for param in smg.get_all_params():
                result_dict[param] = True
        return result_dict.keys()


    def axis_dialog(self, axis_name):
        dlg = QDialog()
        dlg.setWindowTitle('Select ' + axis_name + ' axis parameter')
        layout = QVBoxLayout()

        listwidget = QListWidget()

        if axis_name=='color':
            listwidget.addItems(['list'])

        listwidget.addItems(self.get_all_params())

        layout.addWidget(listwidget)

        def btn_clicked():
            self.change_axis_param(axis_name, listwidget.currentItem().text())
            dlg.close()

        btn = QPushButton('set new axis parameter')
        btn.clicked.connect(btn_clicked)

        layout.addWidget(btn)
        dlg.setLayout(layout)
        dlg.resize(300, 300)
        dlg.exec()


    def initialize_plot(self):
        self.setBackground((255, 255, 255))

        #if tooltip_message is not None:
        #    self.ci.setToolTip(tooltip_message)

        self.clicked_generation = -1
        self.clicked_score = -1
        self.clicked_sm = None
        self.clicked_smg = ''

        self.plot = self.addPlot(row=0, col=0)#, axisItems=axisItems
        #self.plot.addLegend()

        #tab.item = pg.FillBetweenItem(curve1=tab.curves[1], curve2=tab.curves[2], brush=(255, 0, 0, 100))
        #tab.plot.addItem(tab.item)

        # trendline
        self.mean_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.mean_line)

        # minmax area
        self.max_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.max_line)

        self.min_line = pg.PlotCurveItem([],pen=(0,0,0,255))
        self.plot.addItem(self.min_line)

        self.min_max_fill = pg.FillBetweenItem(curve1=self.max_line, curve2=self.min_line, brush=(0, 0, 0, 30))
        self.plot.addItem(self.min_max_fill)

        #on click highlight scatter item
        self.scatter2 = pg.ScatterPlotItem(size=10, brush=pg.mkBrush(0, 0, 255, 255))
        self.plot.addItem(self.scatter2)
        self.scatter2.sigClicked.connect(self.scatter_double_clicked)


        self.plot.getAxis('bottom').setLabel(text=self.default_x)
        self.plot.getAxis('left').setLabel(text=self.default_y)

        def bottom_axis_clicked(ev):
            self.axis_dialog('bottom')

        self.plot.getAxis('bottom').mouseClickEvent = bottom_axis_clicked

        def left_axis_clicked(ev):
            self.axis_dialog('left')

        self.plot.getAxis('left').mouseClickEvent = left_axis_clicked




    def find_same_path_smg(self, smg):
        found = None
        for attached_smg in self.storage_manager_groups:
            if smg.absolute_path == attached_smg.absolute_path:
                found = attached_smg

        return found

    def remove_StorageManagerGroup(self, smg):
        found = self.find_same_path_smg(smg)
        if found is not None:
            self.storage_manager_groups.remove(found)
            self.plot.removeItem(found.scatter)
            self.plot.removeItem(found.error_bar)
            self.update_indices()
            self.scatter2.clear()

    def get_smg(self, tag):
        for smg in self.storage_manager_groups:
            if smg.Tag == tag:
                return smg

    def add_StorageManagerGroup(self, smg):

        found = self.find_same_path_smg(smg)

        if found is None:
            smg.error_bar = pg.ErrorBarItem()
            smg.scatter = pg.ScatterPlotItem(size=10, name=smg.Tag)#, brush=pg.mkBrush(np.random.randint(0,255), np.random.randint(0,255), 255, 120)
            smg.scatter.sigClicked.connect(self.scatter_clicked)
            #self.plot.legend.addItem(smg.scatter, smg.Tag)
            self.plot.addItem(smg.scatter)
            self.plot.addItem(smg.error_bar)
        else:
            smg.error_bar = found.error_bar
            smg.scatter = found.scatter
            smg.color = found.color
            self.storage_manager_groups.remove(found)

        if not hasattr(smg, 'color'):
            smg.color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))

            #print('ck', smg.color[0], smg.color[1], smg.color[2], 120)
            #color = pg.mkBrush(smg.color[0], smg.color[1], smg.color[2], 120)
            #color = pg.mkBrush(np.random.randint(0, 255), np.random.randint(0, 255), 255, 120)
            #smg.scatter.setData(brush=color)

        self.storage_manager_groups.append(smg)

        self.update_indices()

        self.plot.removeItem(self.scatter2)#bring to front
        self.plot.addItem(self.scatter2)
        self.scatter2.clear()


    def update_indices(self):
        return
        #for i, smg in enumerate(self.storage_manager_groups):
        #    smg.add_virtual_multi_parameter('index', i)

    def add_mean_and_variance(self):
        if self.default_x=='index':
            result_lists = smg.get_multi_param_list([self.default_x, self.default_y], remove_None=True).astype(np.float64)

            mx, my = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.average(a)')
            vx, vy = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.var(a)')




    def add_trendline(self, smg):
        if self.default_x=='gen':
            result_lists = smg.get_multi_param_list([self.default_x, self.default_y], remove_None=True).astype(np.float64)

            xa, ya = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.average(a)')
            xs, ymi = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.min(a)')
            xs, yma = smg.remove_duplicates_get_eval(result_lists[-2], result_lists[-1], evalstr='np.max(a)')

            self.mean_line.setData(xa, ya)
            self.min_line.setData(xa, ymi)
            self.max_line.setData(xa, yma)

            self.mean_line.setVisible(True)
            self.min_line.setVisible(True)
            self.max_line.setVisible(True)
            self.min_max_fill.setVisible(True)
        else:
            self.mean_line.setVisible(False)
            self.min_line.setVisible(False)
            self.max_line.setVisible(False)
            self.min_max_fill.setVisible(False)

    def refresh_data(self):
        single_smg = len(self.storage_manager_groups)==1

        #refresh and get global min max color
        color_vals = []
        for smg in self.storage_manager_groups:
            if self.coloration_param != 'list' and self.global_coloration:
                color_vals += smg.get_param_list(self.coloration_param, remove_None=True)

            smg.refresh()

        #actual refresh
        for smg in self.storage_manager_groups:

            params=['#SM#', self.default_x, self.default_y]

            if self.coloration_param != 'list':
                params += [self.coloration_param]

            data = smg.get_multi_param_list(params, remove_None=True)

            if self.coloration_param == 'list':
                c = pg.mkBrush(smg.color[0], smg.color[1], smg.color[2], 120)
            else:
                if self.global_coloration:
                    min=np.min(color_vals)
                    max=np.max(color_vals)
                else:
                    min=np.min(data[3])
                    max=np.max(data[3])
                d=max-min
                if d==0:
                    d=1
                c=[pg.mkBrush(255-255/d*(c-min), 255/d*(c-min), 0, 120) for c in data[3]]

            smg.scatter.setData(x=data[1], y=data[2], brush=c)  #

            #error = pg.ErrorBarItem(x=x, y=y, top=top, bottom=bottom, beam=0.5)
            if self.default_x == 'index':
                meany= np.mean(data[2])
                stdy = np.std(data[2])
                smg.error_bar.setData(x=np.mean(data[1]), y=meany, top=stdy, bottom=stdy, beam=0.3)
                smg.error_bar.setVisible(True)
            else:
                smg.error_bar.setVisible(False)

            for i, d in enumerate(smg.scatter.data):  # set ids to each point (d[3] very ugly coding...) (each point is a set, not an object)
                d[3] = data[0][i]
                d[4] = smg

        self.scatter2.clear()

        if single_smg:
            self.add_trendline(self.storage_manager_groups[0])






from PymoNNto.Exploration.Evolution.common_UI import *
from PymoNNto.Exploration.Evolution.PlotQTObjects import *

class UI_Plot_Manager(Execution_Manager_UI_Base):

    def get_folder(self):
        return 'Plot_Project_Clones'

    def get_title(self):
        return 'Plot Manager'

    def add_ui_elements(self):

        self.qte=QTextEdit()
        self.qte.setPlainText("""for _ in range(#runs#):
	file_exec({})


###help###
#file_exec({'gene1': 1, ...})
""")
        self.Add_element(self.qte, sidebar=True)


        horizontal_layout = QHBoxLayout()
        self.sidebar_current_vertical_layout.addLayout(horizontal_layout)

        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)

        vertical_layout.addWidget(QLabel('slave file'))
        #vertical_layout.addWidget(QLabel('thread number'))
        vertical_layout.addWidget(QLabel('python cmd'))
        vertical_layout.addWidget(QLabel('runs'))

        self.slave_file_edit = QLineEdit('Exploration/Evolution/slave_example.py')
        #self.thread_number_edit = QLineEdit('4')
        self.python_cmd_edit = QLineEdit("python")
        self.run_count_edit = QLineEdit('10')

        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)
        vertical_layout.addWidget(self.slave_file_edit)
        #vertical_layout.addWidget(self.thread_number_edit)
        vertical_layout.addWidget(self.python_cmd_edit)
        vertical_layout.addWidget(self.run_count_edit)


    def On_Tab_Changed(self, i):
        if i is not None:
            tab = self.tabs.currentWidget()
            if hasattr(tab, 'ssm'):
                self.update_status(tab, tab.name)
                ssm = self.tabs.currentWidget().ssm
                self.set_text(ssm, self.task_name_edit, 'name')
                self.set_text(ssm, self.slave_file_edit, 'slave_file')
                #self.set_text(ssm, self.thread_number_edit, 'thread_number')
                self.set_text(ssm, self.run_count_edit, 'run_count')
                self.set_text(ssm, self.python_cmd_edit, 'python_cmd')

    def get_help_txt(self):
        return '''Every tab corresponds to a folder in:
        Data/Plot_Project_Clones/

        For remote SSH servers the local folder only contains some config files while the project copy is located in the user folder of the server.

        When the evolution is started on a ssh server it creates a "screen" session which is named after the evolution name.

        Useful functions are:

        "screen -list" to show the screen sessions and their ids,
        "screen -r id" to connect to a session where "id" has to be replaced and
        "exit" to close a screen session you are connected to.

        The project folders also contain the file execute.py which is used to start the evolution.

        All genomes generate a file with the results and the genes when set_score is called in the slave file. Which are transferred to the local folder when the state is refreshed.
        '''

    def create_execution_file(self, name, folder, file):

        #from PymoNNto.Exploration.Evolution.Interface_Functions import *

        exec_file = """from PymoNNto.Exploration.Evolution.Interface_Functions import *

ids = StorageManagerGroup('evo_name').get_param_list('id', remove_None=True)
if len(ids) == 0:
    execution_counter_ = 0
else:
    execution_counter_ = np.max(ids) + 1

def file_exec(genes):
    global execution_counter_
    genes['evo_name']='#name#'
    genes['id'] = execution_counter_
    execute_local_file('#file#', genes)
    execution_counter_+=1
    
"""+self.qte.toPlainText()
#        """
#from PymoNNto.Exploration.Evolution.Interface_Functions import *
#import sys
#
#for run in range(#run_count#):
#    execute_local_file('#slave_file#', {'evo_name': '#name#', 'gen': run, 'id': run})
#        """

        print('generate execute.py...')
        exec_file = exec_file.replace('#name#', name)
        exec_file = exec_file.replace('#file#', self.slave_file_edit.text())
        #exec_file = exec_file.replace('#thread_number#', self.thread_number_edit.text())
        exec_file = exec_file.replace('#runs#', self.run_count_edit.text())

        md_file = open(file, "w")
        md_file.write(exec_file)
        md_file.close()
        return True

    def save_configuration(self, ssm):#SimpleStorageManager
        ssm.save_param('slave_file', self.slave_file_edit.text())
        #ssm.save_param('thread_number', self.thread_number_edit.text())
        ssm.save_param('run_count', self.run_count_edit.text())
        ssm.save_param('python_cmd', self.python_cmd_edit.text())


    def valid_configuration(self):
        return True

    def add_additional_tab_elements(self, tab, name):
        self.Next_H_Block(stretch=10)

        #tab.plot = self.Add_plot(title='results')

        tab.interactive_scatter = self.Add_element(InteractiveScatter(default_x='id', default_y='score'))

        #add_evolution_plot_items(self, tab)
        #tab.folder = get_epc_folder(self.folder) + '/' + name + '/'

    def refresh_view(self, tab):
        #tab.plot scatter...

        smg = StorageManagerGroup(tab.name, data_folder=get_data_folder() + '/' + self.folder + '/' + tab.name + '/Data')
        tab.interactive_scatter.add_StorageManagerGroup(smg)
        tab.interactive_scatter.refresh_data()

        print(smg.get_param_list('score'))

########################################################### Exception handling


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook

if __name__ == '__main__':
    UI_Plot_Manager().show()



from PymoNNto import *

class Classifier_base(AnalysisModule):

    def initialize(self, neurons):
        self.add_tag('classifier')
        self.add_tag('cluster_matrix_classifier')
        self.corrMatrices={}

    def get_cluster_matrix(self, key):
        if key in self.corrMatrices:
            classification = self.get_results()[key]#self.last_call_result()
            idx = np.argsort(classification)
            return self.corrMatrices[key].iloc[idx, :].T.iloc[idx, :], idx #[idx, :].T.iloc[idx, :]
        else:
            print('module has to be executed first.')

    def get_data_matrix(self, neurons):
        return #overrride

    def execute(self, neurons, sensitivity=2):
        print('computing cluster classes...')
        import scipy.cluster.hierarchy as sch
        import pandas as pd

        data = self.get_data_matrix(neurons)

        mask = np.sum(data, axis=1) > 0
        #mask = np.sum(data, axis=0) > 0#test!
        self.update_progress(10)

        df = pd.DataFrame(data[mask])  # .T
        #df = pd.DataFrame(data[:, mask])  # .T#test!
        self.corrMatrix = df.corr()
        self.update_progress(40)

        self.corrMatrices[self.current_key] = self.corrMatrix

        pairwise_distances = sch.distance.pdist(self.corrMatrix)
        self.update_progress(60)
        linkage = sch.linkage(pairwise_distances, method='complete')
        self.update_progress(70)
        cluster_distance_threshold = pairwise_distances.max() / sensitivity
        idx_to_cluster_array = sch.fcluster(linkage, cluster_distance_threshold, criterion='distance')
        self.update_progress(90)

        result = np.zeros(data.shape[0]) - 1
        #result = np.zeros(data.shape[1]) - 1#test!

        print(result.shape, mask, idx_to_cluster_array)
        result = idx_to_cluster_array
        #result[mask] = idx_to_cluster_array

        #match_groups
        if self.current_key in self.result_storage:
            old = self.result_storage[self.current_key]
            new = result.copy()

            oc = np.max(old) + 1
            nc = np.max(new) + 1
            mc = np.maximum(oc, nc)
            map_matrix = np.zeros((int(mc), int(mc)))

            for c in np.unique(old):
                old_c_indx = np.where(old == c)[0]
                values = new[old_c_indx]
                for v in values:
                    map_matrix[int(c), int(v)] += 1

            mapping = {}
            for _ in range(int(nc)):
                oi, ni = np.unravel_index(map_matrix.argmax(), map_matrix.shape)
                map_matrix[oi, :] = -1
                map_matrix[:, ni] = -1
                mapping[ni] = oi
                result[new == ni] = oi
            print('applied class mapping from old state (new_c:old_c)', mapping)
        return result #classification is stored/updated in the results list

'''
old = np.array([2,3,2,1,1,0])
new = np.array([1,0,1,2,1,0])

oc = np.max(old) + 1
nc = np.max(new) + 1
mc = np.maximum(oc, nc)
map_matrix = np.zeros((int(mc), int(mc)))

for c in np.unique(old):
    old_c_indx = np.where(old == c)[0]
    values = new[old_c_indx]
    for v in values:
        map_matrix[int(c), int(v)] += 1

result = np.zeros(new.shape)

mapping = {}
for _ in range(int(nc)):
    oi, ni = np.unravel_index(map_matrix.argmax(), map_matrix.shape)
    print(oi, ni)
    map_matrix[oi, :] = -1
    map_matrix[:, ni] = -1
    mapping[ni] = oi
    result[new == ni] = oi

print(result)
print(mapping)
'''






from PymoNNto.Exploration.StorageManager.StorageManager import *
import sys

evolution_genome = None
default_genome = {}

###############################old

#def load_genome2():
#    for arg in sys.argv:
#        if 'genome=' in arg:
#            set_evolution_genome_from_string(arg[7:])

#def set_evolution_genome_from_string(gene_str):
#    #print('gene_str', gene_str)
#    gene_dict = {}
#    partial_str = gene_str.split('#')
#    for key_value_str in partial_str:
#        kv_partial_str = key_value_str.split('@')
#        if len(kv_partial_str) == 2:
#            if kv_partial_str[0] != 'score':
#                gene_dict[kv_partial_str[0]] = kv_partial_str[1] #keys and values are strings!!! have to be casted to default type by get_gene!
#
#    set_genome(gene_dict)

#######################################

def execute_local_file(file, genome):
    py_file = open(file, "r")
    file_content = py_file.read()
    py_file.close()
    for arg in sys.argv:
        if '=' in arg:
            sys.argv.remove(arg)
    for gene in genome:
        sys.argv.append(str(gene) + '=' + str(genome[gene]))
    exec(file_content)

def load_genome():# pythonfile.py gene1=5 gene2=a gene3=...
    gene_dict = {}
    for arg in sys.argv[1:]:
        if '=' in arg:
            key_value = arg.split('=')
            if key_value[0]!='score':
                gene_dict[key_value[0]]=arg.replace(key_value[0]+'=','')
    #print(gene_dict)
    set_genome(gene_dict)

def set_genome(genome):
    #print('set genome to', genome)
    global evolution_genome
    evolution_genome = genome
    #if not 'evo_name' in genome:
    #    print('warning: no evo_name key in evolution for', genome)

def get_genome():
    if evolution_genome is None:
        load_genome()
    return evolution_genome

def cast_to_default(str_value, default):
    if isinstance(default, float) or isinstance(default, int):
        return float(str_value)
    else:
        return str_value

def get_gene(key, default):
    if evolution_genome is None:
        load_genome()

    default_genome[key] = default
    if evolution_genome is not None and key in evolution_genome:
        return cast_to_default(evolution_genome[key], default)
    else:
        return default

def get_default_genome():
    return default_genome

def gene(key, default):
    return get_gene(key, default)

#def get_gene_id(gene):
#    id = ''
#    for key, value in gene.items():
#        id += '#'+key+'@'+str(value)
#    return id+'#'

def get_gene_file(gene):
    file = gene['evo_name']
    if 'gen' in  gene:
        file+=' gen' + str(gene['gen'])
    if 'id' in gene:
        file += ' id' + str(gene['id'])
    return file

def set_score(score, non_evo_storage_manager=None, _genome=None, info={}):
    global evolution_genome

    if _genome is not None:
        evolution_genome = _genome

    if 'evo_name' in evolution_genome:#evolution_genome is not None
        if 'score' in evolution_genome:#only when setscore is called multiple times on accident...
            evolution_genome.pop('score')
        #if 'evo_name' in evolution_genome:# and 'gen' in evolution_genome and 'id' in evolution_genome
        sm = StorageManager(main_folder_name=get_gene('evo_name', None), folder_name=get_gene_file(evolution_genome), print_msg=False, add_new_when_exists=False)
        evolution_genome['score'] = score
        sm.save_param_dict(evolution_genome)
        sm.save_param_dict(info)
        print('evolution score set to #'+str(score)+'#')
        #else:
        #    print('cannot save score', str(score), 'to file: "evo_name", "gen" or "id" not in genome')
    else:
        print('score='+str(score)+' (no genome found)')
        if non_evo_storage_manager is not None:
            non_evo_storage_manager.save_param('score', score)
            non_evo_storage_manager.save_param_dict(info)

    # reset and reload when next get_gene is called

    evolution_genome = None






from PymoNNto.Exploration.Evolution.common_UI import *

class UI_Evolution_Manager(Execution_Manager_UI_Base):

    def get_folder(self):
        return 'Evolution_Project_Clones'

    def get_title(self):
        return 'Evolution Monitor'

    def add_ui_elements(self):
        horizontal_layout = QHBoxLayout()
        self.sidebar_current_vertical_layout.addLayout(horizontal_layout)

        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)
        vertical_layout.addWidget(QLabel('slave file'))
        vertical_layout.addWidget(QLabel('thread number'))
        vertical_layout.addWidget(QLabel('python cmd'))
        vertical_layout.addWidget(QLabel('individual count'))
        vertical_layout.addWidget(QLabel('mutation'))
        vertical_layout.addWidget(QLabel('death rate'))
        sg=QLabel('start geneomes')
        vertical_layout.addWidget(sg)
        sg.setFixedHeight(200)
        vertical_layout.addWidget(QLabel('inactive info'))
        vertical_layout.addWidget(QLabel('constriants'))
        vertical_layout.addWidget(QLabel('evo options'))

        self.slave_file_edit = QLineEdit('Exploration/Evolution/slave_example.py')
        self.thread_number_edit = QLineEdit('4')
        self.python_cmd_edit = QLineEdit("python3")
        self.individual_count_edit = QLineEdit('10')
        self.mutation_edit = QLineEdit('0.05')
        self.death_rate_edit = QLineEdit('0.5')
        self.start_genomes_edit = QTextEdit("[{'a':1,'b':2,'c':2,'d':2,'e':3}]")
        self.start_genomes_edit.setFixedHeight(200)
        #QLineEdit("[{'a':1,'b':2,'c':2,'d':2,'e':3}]")
        self.inactive_genome_info_edit = QLineEdit('{}')
        self.constraints_edit = QLineEdit("['a<b','b<=c']")
        self.evo_options_edit = QLineEdit('{}')

        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)
        vertical_layout.addWidget(self.slave_file_edit)
        vertical_layout.addWidget(self.thread_number_edit)
        vertical_layout.addWidget(self.python_cmd_edit)
        vertical_layout.addWidget(self.individual_count_edit)
        vertical_layout.addWidget(self.mutation_edit)
        vertical_layout.addWidget(self.death_rate_edit)
        vertical_layout.addWidget(self.start_genomes_edit,stretch=5)
        vertical_layout.addWidget(self.inactive_genome_info_edit)
        vertical_layout.addWidget(self.constraints_edit)
        vertical_layout.addWidget(self.evo_options_edit)

    def On_Tab_Changed(self, i):
        if i is not None:
            tab = self.tabs.currentWidget()
            if hasattr(tab, 'ssm'):
                self.update_status(tab, tab.name)
                ssm = self.tabs.currentWidget().ssm
                self.set_text(ssm, self.task_name_edit, 'name')
                self.set_text(ssm, self.slave_file_edit, 'slave_file')
                self.set_text(ssm, self.inactive_genome_info_edit, 'inactive_genome_info')
                self.set_text(ssm, self.thread_number_edit, 'thread_number')
                self.set_text(ssm, self.individual_count_edit, 'individual_count')
                self.set_text(ssm, self.mutation_edit, 'mutation')
                self.set_text(ssm, self.death_rate_edit, 'death_rate')
                self.set_text(ssm, self.start_genomes_edit, 'start_genomes')
                self.set_text(ssm, self.constraints_edit, 'constraints')
                self.set_text(ssm, self.python_cmd_edit, 'python_cmd')
                self.set_text(ssm, self.evo_options_edit, 'evo_options')

    def get_help_txt(self):
        return '''Every tab corresponds to a folder in:
        Data/Evolution_Project_Clones/

        For remote SSH servers the local folder only contains some config files while the project copy is located in the user folder of the server.

        When the evolution is started on a ssh server it creates a "screen" session which is named after the evolution name.

        Useful functions are:

        "screen -list" to show the screen sessions and their ids,
        "screen -r id" to connect to a session where "id" has to be replaced and
        "exit" to close a screen session you are connected to.

        The project folders also contain the file execute.py which is used to start the evolution.

        All genomes generate a file with the results and the genes when set_score is called in the slave file. Which are transferred to the local folder when the state is refreshed.
        '''

    def create_execution_file(self, name, folder, file):
        exec_file = """
from PymoNNto.Exploration.Evolution.Evolution import *
if __name__ == '__main__':
    evo = Evolution(name='#name#',
                slave_file='#slave_file#',
                individual_count=#individual_count#,
                mutation=#mutation#,
                death_rate=#death_rate#,
                constraints=#constraints#,
                inactive_genome_info=#inactive_genome_info#,
                start_genomes=#start_genomes#,
                devices={'multi_thread': #thread_number#},
                additional_evo_params=#evo_options#
                )

    if not evo.start(ui=False):
        evo.continue_evolution(ui=False)
        """

        print('generate execute.py...')
        exec_file = exec_file.replace('#name#', name)
        exec_file = exec_file.replace('#slave_file#', self.slave_file_edit.text())
        exec_file = exec_file.replace('#inactive_genome_info#', self.inactive_genome_info_edit.text())
        exec_file = exec_file.replace('#thread_number#', self.thread_number_edit.text())
        exec_file = exec_file.replace('#individual_count#', self.individual_count_edit.text())
        exec_file = exec_file.replace('#mutation#', self.mutation_edit.text())
        exec_file = exec_file.replace('#death_rate#', self.death_rate_edit.text())
        exec_file = exec_file.replace('#start_genomes#', self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))
        exec_file = exec_file.replace('#constraints#', self.constraints_edit.text())
        exec_file = exec_file.replace('#evo_options#', self.evo_options_edit.text())

        md_file = open(file, "w")
        md_file.write(exec_file)
        md_file.close()
        return True

    def save_configuration(self, ssm):#SimpleStorageManager

        ssm.save_param('slave_file', self.slave_file_edit.text())
        ssm.save_param('inactive_genome_info', self.inactive_genome_info_edit.text())
        ssm.save_param('thread_number', self.thread_number_edit.text())
        ssm.save_param('individual_count', self.individual_count_edit.text())
        ssm.save_param('mutation', self.mutation_edit.text())
        ssm.save_param('death_rate', self.death_rate_edit.text())
        ssm.save_param('start_genomes', self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))
        ssm.save_param('constraints', self.constraints_edit.text())
        ssm.save_param('python_cmd', self.python_cmd_edit.text())
        ssm.save_param('evo_options', self.evo_options_edit.text())

        gene_keys = list(eval(self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))[0].keys())
        ssm.save_param('gene_keys', gene_keys)

    def valid_configuration(self):
        valid_genomes = False
        try:
            sg = eval(self.start_genomes_edit.toPlainText().replace('\r', '').replace('\n', ''))
            for key in ['gen', 'score', 'id']:
                for genome in sg:
                    if key in genome:
                        genome.pop(key)
                        print(key, 'removed from genomes')
            self.start_genomes_edit.setText(str(sg))
            valid_genomes = True
        except:
            print('error parsing start genomes')

        if not valid_genomes:
            print('error parsing start genomes')

        return valid_genomes

    def add_additional_tab_elements(self, tab, name):
        self.Next_H_Block(stretch=10)
        add_evolution_plot_items(self, tab)
        tab.folder = get_epc_folder(self.folder) + '/' + name + '/'

    def refresh_view(self, tab):
        if tab.gene_keys is not None:
            update_evolution_plot(self, tab, tab.name, tab.gene_keys,data_folder=get_data_folder() + '/' + self.folder + '/' + tab.name + '/Data')

########################################################### Exception handling


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook

if __name__ == '__main__':
    UI_Evolution_Manager().show()




import random
import numpy as np
from PymoNNto.Exploration.Evolution.Interface_Functions import *

class Default_Breed_And_Select:

    def __init__(self, parent, death_rate, mutation, individual_count, constraints, additional_evo_params={}):
        self.parent = parent
        self.death_rate = death_rate
        self.mutation = mutation
        self.individual_count = individual_count
        self.generation = 1
        self.constraints = constraints

        if 'no_high_risk_inds' in additional_evo_params:
            self.avoid_high_risk_high_reward_individuals = additional_evo_params['no_high_risk_inds']
        else:
            self.avoid_high_risk_high_reward_individuals = True

        if 're_eval_inds' in additional_evo_params:
            self.re_evaluate_genomes = additional_evo_params['re_eval_inds']
        else:
            self.re_evaluate_genomes = True


    def update_population(self):
        if len(self.parent.non_scored_individuals) == 0 and len(self.parent.running_individuals) == 0:

            survivours = self.natural_selection(self.parent.scored_individuals)

            if self.avoid_high_risk_high_reward_individuals:
                for s in survivours:# improve breed chance for older individuals (avoid high risk, high reward population)
                    bc = 1
                    if 'mate_chance' in s:
                        bc = s['mate_chance']
                    s['mate_chance'] = (bc+s['score'])/2

            new_population = self.breed(survivours)

            for p in new_population:#remove scores
                if self.re_evaluate_genomes or p not in survivours:
                    if 'score' in p:
                        p.pop('score')

            print('new gen breeding... survivours', survivours)

            self.parent.non_scored_individuals = new_population.copy()
            self.parent.running_individuals = []
            self.parent.scored_individuals = []
            self.generation += 1

            if not self.re_evaluate_genomes:
                for s in survivours:
                    found = self.parent.find(self.parent.non_scored_individuals, s)
                    self.parent.non_scored_individuals.remove(found)
                    self.parent.running_individuals.append(found)

                    processing_genome = self.parent.add_name_gen_id_inactive_to_genome(s)
                    set_score(s['score'], _genome=processing_genome)#create files
                    self.parent.new_score_event(processing_genome)#move from non_scored_individuals to scored_individuals


    def natural_selection(self, individuals):
        result = individuals.copy()
        kill_count = int(len(result) * self.death_rate)

        for _ in range(kill_count):
            kill_ind = result[0]
            for ind in result:
                if ind['score'] < kill_ind['score']:
                    kill_ind = ind
            result.remove(kill_ind)

        #sort (biggest score first) (not neccessary, just for better overview)
        for i in range(len(result)-1):
            for j in range(len(result)-i-1):
                if result[i]['score'] < result[i+1+j]['score']:
                    temp = result[i]
                    result[i] = result[i+1+j]
                    result[i+1+j] = temp

        return result

    def is_valid_genome(self, genome):
        if genome is None:
            return False
        else:
            result = True
            for gene_key, gene_value in genome.items():
                locals()[gene_key] = gene_value

            for constraint in self.constraints:
                result = result and eval(constraint)

            return result


    def get_offspring(self, ind1, ind2):
        genome = None
        while not self.is_valid_genome(genome):
            genome = {}

            for key in ind1.keys():
                if key != 'mate_chance' and key != 'score':
                    if key in ind2:
                        g1 = np.random.normal(ind1[key], np.clip(ind1[key] * self.mutation, 0, None))
                        g2 = np.random.normal(ind2[key], np.clip(ind2[key] * self.mutation, 0, None))
                        genome[key] = random.choice([g1, g2])
                    else:
                        print('not the same genes', ind1, ind2)

                genome['mate_chance'] = 1.0

        return genome


    def breed(self, individuals):

        new_count = self.individual_count-len(individuals)

        result = []
        for ind in individuals:
            result.append(ind.copy())

        weights = np.ones(len(individuals))
        for i, ind in enumerate(individuals):
            if 'mate_chance' in ind:
                weights[i] = ind['mate_chance']

        weights = weights / np.sum(weights)

        for i in range(new_count):
            parent1 = np.random.choice(individuals, 1, p=weights)[0]
            parent2 = np.random.choice(individuals, 1, p=weights)[0]

            result.append(self.get_offspring(parent1, parent2))

        return result




from PymoNNto.NetworkCore.Network import *
from PymoNNto.NetworkCore.Behaviour import *
from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.NetworkCore.Analysis_Module import *

from PymoNNto.Exploration.HelperFunctions import *
from PymoNNto.Exploration.StorageManager.StorageManager import *

import os

class Counter(Behaviour):
    def set_variables(self, neurons):
        self.inc = self.get_init_attr('inc', 1)
        neurons.count = neurons.get_neuron_vec()
    def new_iteration(self, neurons):
        neurons.count += self.inc

v = 1

def test_basics():
    set_genome({'I':1})

    #basic network
    My_Network = Network()
    My_Neurons = NeuronGroup(net=My_Network, tag='my_neurons', size=100, behaviour={
        1: Counter(inc='[2#I]'),
        2: Recorder(variables=['n.count'])
    })
    My_Synapses = SynapseGroup(net=My_Network, src=My_Neurons, dst=My_Neurons, tag='GLUTAMATE')

    sm = StorageManager('test', random_nr=False, print_msg=False)

    My_Network.initialize(storage_manager=sm)
    My_Network.simulate_iterations(1000)

    My_Network.deactivate_mechanisms('Counter')
    My_Network.simulate_iterations(10)

    My_Network.activate_mechanisms('Counter')
    My_Network.simulate_iterations(20)

    My_Network.recording_off()
    My_Network.simulate_iterations(30)

    assert My_Network.iteration == 1000+10+20+30
    assert np.sum(My_Neurons.count) == My_Neurons.size*(1000+20+30)

    assert My_Synapses.src == My_Neurons
    assert My_Synapses.dst == My_Neurons

    assert My_Neurons.afferent_synapses['GLUTAMATE'] == [My_Synapses]

    assert len(My_Network.all_objects()) == 3

    #tagging system
    assert My_Network['my_neurons'] == [My_Neurons]
    assert len(My_Network['n.count', 0]) == My_Network.iteration-30

    My_Network.clear_recorder()
    assert len(My_Neurons['n.count', 0]) == 0

    assert My_Network.tag_shortcuts['my_neurons'] == My_Network['my_neurons']

    #Storage Manager
    assert os.path.isfile('Data/StorageManager/test/test/config.ini')
    sm.save_param('k', 'v')
    assert sm.load_param('k') == 'v'


from PymoNNto.Exploration.UI_Base import *
from PymoNNto.Exploration.StorageManager.StorageManager import *
from PymoNNto.Exploration.Evolution.SSH_Functions import *
from PymoNNto.Exploration.Evolution.EvolutionPlots import *
from multiprocessing import Process, Queue, Pipe
import shutil

class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)

class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)

def evolution_thread_worker(name, folder, python_cmd, conn):
    print('local thread started')
    try:
        dir = os.getcwd()
        os.chdir(get_epc_folder(folder)+'/'+name)
        print('move to', get_epc_folder(folder)+'/'+name)

        system_str = sys.platform
        if 'darwin' in system_str:#mac os
            os.system(python_cmd+' execute.py')
        elif 'win' in system_str:#windows
            os.system('start /wait cmd /c '+python_cmd+' execute.py')
        else:
            os.system('gnome-terminal -x '+python_cmd+' execute.py')

        os.chdir(dir)

        conn.send('execution finished')
    except:
        conn.send('thread error')
    conn.send('idle')


def copy_project(evo_name, device_str, folder):
    if device_str == 'local':
        clone_project(evo_name, folder)

    if 'ssh' in device_str:
        user, host, password = split_ssh_user_host_password_string(device_str)
        if user is not None and host is not None:
            transfer_project(evo_name, user, host, password)

        Data_Folder = get_data_folder()
        if Data_Folder != './Data':
            epc_folder = Data_Folder + '/'+folder
            create_folder_if_not_exist(epc_folder)
            run_folder = epc_folder + '/' + evo_name
            create_folder_if_not_exist(run_folder)


class Execution_Manager_UI_Base(UI_Base):

    def add_ui_elements(self):
        return

    def On_Tab_Changed(self, i):
        return

    def get_help_txt(self):
        return '''...
        '''

    def create_execution_file(self, name, folder, file):
        return False


    def save_configuration(self,ssm):#SimpleStorageManager
        return


    def valid_configuration(self):
        return True

    def add_additional_tab_elements(self, tab, name):
        return

    def get_folder(self):
        return 'Execution_Project_Clones'

    def get_title(self):
        return 'Execute and Transfer'

    def __init__(self):
        super().__init__(None, label=self.get_title(), create_sidebar=True)

        self.folder = self.get_folder()

        self.tabs.currentChanged.connect(self.On_Tab_Changed)


        horizontal_layout = QHBoxLayout()
        self.sidebar_current_vertical_layout.addLayout(horizontal_layout)

        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)
        vertical_layout.addWidget(QLabel('task name'))
        self.task_name_edit = QLineEdit('task_name')
        vertical_layout = QVBoxLayout()
        horizontal_layout.addLayout(vertical_layout)
        vertical_layout.addWidget(self.task_name_edit)



        ssm = SimpleStorageManager(get_epc_folder() + '/')
        servers = eval(ssm.load_param('servers', default="[]"))
        self.listwidget2 = QListWidget()
        self.listwidget2.addItems(['+', 'local'] + servers)
        self.listwidget2.currentItemChanged.connect(self.on_server_select)
        self.Add_element(self.listwidget2, stretch=10, sidebar=True)

        self.add_ui_elements()

        self.add_btn = self.Add_element(QPushButton('Add'), sidebar=True, stretch=1)
        self.add_btn.clicked.connect(self.on_add_click)

        for dir in os.listdir(get_epc_folder(self.folder)):
            if os.path.isdir(get_epc_folder(self.folder) + '/' + dir):
                self.add_tab(dir)

    def keyPressEvent(self, event):
        if event.key() in [Qt.Key_I]:
            tab = self.tabs.currentWidget()
            print(tab.clicked_id)
            smg = StorageManagerGroup(tab.name, data_folder=get_data_folder() + '/'+self.folder+'/' + tab.name + '/Data')
            smg.sort_by('gen')
            sm = smg.StorageManagerList[tab.clicked_id]
            file = sm.absolute_path + sm.config_file_name
            print(file)
            with open(file) as f:
                lines = f.readlines()

                layout = QVBoxLayout()
                pte = QPlainTextEdit()
                pte.setPlainText('\r\n'.join(lines))
                pte.setReadOnly(True)
                layout.addWidget(pte)
                dlg = QDialog()
                dlg.setWindowTitle("Info")
                dlg.setLayout(layout)
                dlg.resize(600, 400)
                dlg.exec()

    def set_text(self, ssm, edit, text):
        data = ssm.load_param(text, default='', return_string=True)
        if data is not None and data != '':
            edit.setText(data)

    def update_status(self, tab, name):
        if 'ssh ' in tab.server:
            if name in ssh_get_running(tab.server):
                tab.status_label.setText('<font color="green">running</font>')
                return True
            else:
                tab.status_label.setText('<font color="red">inactive</font>')
                return False

        if 'local' in tab.server:
            if tab.process is not None:
                if tab.parent_conn.poll():
                    msg = tab.parent_conn.recv()
                    tab.process.kill()
                    tab.process = None
                    tab.status_label.setText('<font color="red">inactive</font>')
                    return False
                else:
                    tab.status_label.setText('<font color="green">running</font>')
                    return True
            else:
                tab.status_label.setText('<font color="red">inactive</font>')
                return False

    def refresh_view(self, tab):
        return

    def add_tab(self, name):

        folder = get_epc_folder(self.folder) + '/' + name + '/'

        tab = self.Next_Tab(name, stretch=0)
        tab.name = name

        tab.ssm = SimpleStorageManager(folder)
        tab.server = tab.ssm.load_param('server', default=None)

        tab.gene_keys = tab.ssm.load_param('gene_keys', default=None)
        if tab.gene_keys is not None:
            tab.gene_keys = eval(tab.gene_keys)

        tab.process = None
        tab.python_cmd = tab.ssm.load_param('python_cmd', default='python3')



        if tab.ssm.load_param('executable', default=None) is not None:###################################################################################################

            tab.status_label = self.Add_element(QLabel('Status'))

            def refresh():
                try:
                    print('refresh')
                    if tab.server is not None and 'ssh ' in tab.server:
                        user, host, password = split_ssh_user_host_password_string(tab.server)
                        print('get Data')
                        get_Data(name, user, host, password, self.folder)

                    self.refresh_view(tab)
                except:
                    print('no plot data found in', name)

                self.update_status(tab, name)

            refresh_btn = self.Add_element(QPushButton('refresh state'))
            refresh_btn.clicked.connect(refresh)

            def start_continue():
                if 'ssh ' in tab.server:
                    if not self.update_status(tab, name):
                        ssh_execute_evo(tab.server, name, python_cmd=tab.python_cmd)  # python3
                    else:
                        print('thread already running')

                if 'local' in tab.server:
                    if tab.process is None:
                        tab.parent_conn, child_conn = Pipe()
                        tab.process = Process(target=evolution_thread_worker,
                                              args=(name, self.folder, tab.python_cmd, child_conn))  # python
                        print('start', self.folder+'/'+name+'/execute.py')
                        tab.process.start()
                    else:
                        print('thread already running')

                self.update_status(tab, name)

            start_evo_btn = self.Add_element(QPushButton('start/continue'))
            start_evo_btn.clicked.connect(start_continue)

            def stop():
                if 'ssh ' in tab.server:
                    ssh_stop_evo(tab.server, name)
                    self.update_status(tab, name)
                if 'local' in tab.server:
                    if tab.process is not None:
                        print('Please close terminal manually')
                        tab.process.kill()
                        tab.process.terminate()
                        tab.process.close()
                        tab.process = None

            stop_btn = self.Add_element(QPushButton('stop evolution'))
            stop_btn.clicked.connect(stop)

        #############################################################################################################

        def remove():
            if 'ssh ' in tab.server:
                ssh_stop_evo(tab.server, name, remove_evo=True)

            if 'local' in tab.server:
                if tab.process is not None:
                    print('Please close terminal manually')
                    tab.process.kill()
                    tab.process.terminate()
                    tab.process.close()
                    tab.process = None

            shutil.rmtree(get_epc_folder(self.folder) + '/' + name + '/')
            self.tabs.removeTab(self.tabs.currentIndex())

        remove_btn = self.Add_element(QPushButton('remove'))
        remove_btn.clicked.connect(remove)

        def help():
            layout = QVBoxLayout()
            pte = QPlainTextEdit()
            pte.setPlainText(self.get_help_txt())
            pte.setReadOnly(True)
            layout.addWidget(pte)

            # self.layout.addWidget(message)
            # self.layout.addWidget(self.buttonBox)

            dlg = QDialog()
            dlg.setWindowTitle("Help")
            dlg.setLayout(layout)
            dlg.resize(600, 400)
            dlg.exec()

        help_btn = self.Add_element(QPushButton('help'))
        help_btn.clicked.connect(help)

        self.Add_element(QLabel(tab.server))

        self.add_additional_tab_elements(tab, name)

        self.On_Tab_Changed(None)


    def on_server_select(self):
        server_str = self.listwidget2.currentItem().text()
        if server_str == '+':
            text, ok = QInputDialog.getText(self.main_window, 'Add...',
                                            'Add computing device "ssh user@host" or "ssh user@host password"')
            if ok and text != '+':
                self.listwidget2.addItem(text)
                ssm = SimpleStorageManager(get_epc_folder(self.folder) + '/')
                print(ssm.absolute_path)
                ssm.save_param('servers', eval(ssm.load_param('servers', default="[]")) + [text])

    def on_file_click(self):
        print(self.listwidget.currentItem().text())

    def on_add_click(self):
        name = self.task_name_edit.text()
        if is_invalid_evo_name(name):
            print(
                'invalid evolution name.\r\n For savety reasons some names and characters are forbidden to avoid the accidental removal of files or folders')
        else:
            item = self.listwidget2.currentItem()
            if item is None:
                print('please select a server first')
            else:
                server_str = item.text()
                if os.path.exists(get_data_folder() + '/'+self.folder+'/' + name):
                    print('folder already exists: '+ get_data_folder() + '/'+self.folder+'/' + name)
                elif self.valid_configuration():

                    copy_project(name, server_str, self.folder)

                    folder = get_epc_folder(self.folder) + '/' + name + '/'

                    ssm = SimpleStorageManager(folder)
                    ssm.save_param('server', server_str)
                    ssm.save_param('name', name)

                    self.save_configuration(ssm)

                    evo_file = folder + 'execute.py'
                    if self.create_execution_file(name, folder, evo_file):
                        ssm.save_param('executable', 'True')

                    print('saving config...')

                    if 'ssh' in server_str:
                        user, host, password = split_ssh_user_host_password_string(server_str)
                        print('transfer execute.py to ' + host + '...')
                        ssh = get_ssh_connection(host, user, password)
                        scp = SCPClient(ssh.get_transport())
                        scp.put(evo_file, name + '/execute.py')
                        scp.close()
                        ssh.close()

                    print('adding tab...')
                    self.add_tab(name)

if __name__ == '__main__':
    Execution_Manager_UI_Base().show()


from PymoNNto.Exploration.Network_UI.TabBase import *

class sidebar_module_quick_access(TabBase):

    def __init__(self, module_tags=[]):
        self.module_tags = module_tags

    def module_on_off(self, cb, Network_UI):
        module_tag=cb.module_tag
        print(module_tag)
        Network_UI.network.set_mechanisms([module_tag], cb.isChecked())
        Network_UI.add_event(module_tag + ' (' + str(cb.isChecked()) + ')')

    def add_cb(self, module_tag, Network_UI):

        def module_on_off(event):
            print(module_tag)
            Network_UI.network.set_mechanisms([module_tag], cb.isChecked())
            Network_UI.add_event(module_tag + ' (' + str(cb.isChecked()) + ')')

        cb = QCheckBox()
        cb.module_tag = module_tag
        cb.setText(module_tag)
        cb.setChecked(Network_UI.network[module_tag, 0].behaviour_enabled)
        cb.stateChanged.connect(module_on_off)
        Network_UI.Add_Sidebar_Element(cb)

    def initialize(self, Network_UI):

        for module_tag in self.module_tags:
            if Network_UI.network.has_module(module_tag):# Network_UI.network[self.module_tag, 0] is not None:
                self.add_cb(module_tag, Network_UI)

        '''
        if Network_UI.network['STDP', 0] is not None:
            def learning_on_off(event):
                Network_UI.network.set_mechanisms(['STDP'], self.stdp_cb.isChecked())
                Network_UI.add_event('STDP (' + str(self.stdp_cb.isChecked()) + ')')

            self.stdp_cb = QCheckBox()
            self.stdp_cb.setText('STDP')
            self.stdp_cb.setChecked(Network_UI.network['STDP', 0].behaviour_enabled)
            self.stdp_cb.stateChanged.connect(learning_on_off)
            Network_UI.Add_Sidebar_Element(self.stdp_cb)

        if Network_UI.network['Text_Activator', 0] is not None:
            def activator_on_off(event):
                Network_UI.network['Text_Activator', 0].behaviour_enabled = self.act_cb.isChecked()
                Network_UI.add_event('Text_Activator (' + str(self.act_cb.isChecked()) + ')')

            self.act_cb = QCheckBox()
            self.act_cb.setText('Text input:')
            self.act_cb.setChecked(Network_UI.network['Text_Activator', 0].behaviour_enabled)
            self.act_cb.stateChanged.connect(activator_on_off)
            Network_UI.Add_Sidebar_Element(self.act_cb)
        '''


from PymoNNto.Exploration.AnalysisModules.Weight_Classifier_Base import *

class Weight_Classifier_Post(Classifier_base):

    def get_data_matrix(self, neurons):
        syn_tag = self.get_init_attr('syn_tag', 'EE')
        return get_partitioned_synapse_matrix(neurons, syn_tag, 'W')



from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Visualization_Helper import *
from PymoNNto.Exploration.Network_UI.Network_UI import *

class classifier_bar_tab(TabBase):


    def initialize(self, Network_UI):
        self.cluster_bar_tab = Network_UI.Next_Tab('Cluster')
        self.plot = Network_UI.Add_plot('cluster sizes')
        self.bgi = pg.BarGraphItem(x=[], height=[], width=0.6, brush='r')
        self.plot.addItem(self.bgi)

        self.class_id_positions = {}


        def mouseClickEvent(event):
            click_y_pos = np.round(event.pos().y())
            for c in self.class_id_positions:
                y_pos = self.class_id_positions[c]
                if click_y_pos == y_pos:
                    Network_UI.select_neuron_class(self.current_group, c)

        self.bgi.mouseClickEvent = mouseClickEvent



        Network_UI.Next_H_Block()

        #self.vertical_cb = Network_UI.Add_element(QCheckBox('vertical'))

        self.class_tag_dict = None

        def on_new_label_selected(group, module, key, labels):
            generator = group.network['Text_Generator', 0]
            if hasattr(group, 'classification') and module is not None and generator is not None:
                self.class_tag_dict = module.get_class_labels(key, group.classification, generator)

        self.comboBox = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], tag='labeler', first_entry='none'))
        self.comboBox.on_update_func = on_new_label_selected


    def update(self, Network_UI):
        if self.cluster_bar_tab.isVisible():
            self.current_group = Network_UI.selected_neuron_group()
            if hasattr(self.current_group, 'classification'):
                selected_classes = Network_UI.selected_class_ids()

                idx = []
                num = []
                col = []
                labels = []
                cd = self.current_group['Neuron_Classification_Colorizer', 0].get_class_color_dict(self.current_group.classification, format='[r,g,b]')#.get_color_list(group.classification, format='[r,g,b]'))

                for i, c in enumerate(np.unique(self.current_group.classification)):

                    idx.append(i)
                    num.append(np.sum(self.current_group.classification == c))

                    if c in selected_classes:
                        col.append((0, 255, 0))
                    else:
                        col.append(cd[c])

                    if self.class_tag_dict is not None and c in self.class_tag_dict:
                        labels.append(self.class_tag_dict[c])

                sort_indx = np.argsort(-np.array(num))


                for i, c in enumerate(np.unique(self.current_group.classification)[sort_indx]):
                    self.class_id_positions[c] = i
                    #if self.clicked_y_pos == i:
                    #    self.clicked_y_pos = -1


                #if self.vertical_cb.isChecked():
                self.bgi.setOpts(x=np.array(num)[sort_indx] / 2, y=idx, height=0.6, width=np.array(num)[sort_indx], brushes=np.array(col)[sort_indx])
                #else:
                #    self.bgi.setOpts(x=idx, y=np.array(num)[sort_indx] / 2, height=np.array(num)[sort_indx], width=0.6, brushes=np.array(col)[sort_indx])

                #if len(labels) > 0:
                    #if self.current_label_resort is not None:
                    #    labels = list(np.flip(np.array(labels)[self.current_label_resort]))
                if len(sort_indx)==len(labels):
                    labels = [(i, l) for i, l in enumerate(np.array(labels)[sort_indx])]

                self.plot.showAxis('left')
                ax = self.plot.getAxis('left')
                ax.setTickFont(QFont("Courier"))
                ax.setWidth(100)
                ax.setTicks([labels])




'''
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui

window = pg.plot()

y1 = [5, 5, 7, 10, 3, 8, 9, 1, 6, 2]
x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
bargraph = pg.BarGraphItem(x=np.array(y1)/2, y=x, width=y1, height=0.6)
window.addItem(bargraph)
QtGui.QApplication.instance().exec_()
'''


from PymoNNto.Exploration.Network_UI.Network_UI import *
from PymoNNto.Exploration.Network_UI.TabBase import *
import scipy.cluster.hierarchy as sch
import pandas as pd
import matplotlib.pyplot as plt

def generatePgColormap(cm_name):
    pltMap = plt.get_cmap(cm_name)
    colors = pltMap.colors
    colors = [c + [1.] for c in colors]
    positions = np.linspace(0, 1, len(colors))
    #print(positions, colors)
    pgMap = pg.ColorMap(positions, np.array(colors)*255.0)
    return pgMap


'''
def cluster_corr(corr_array, inplace=False):
    pairwise_distances = sch.distance.pdist(corr_array)
    linkage = sch.linkage(pairwise_distances, method='complete')
    cluster_distance_threshold = pairwise_distances.max() / 2
    idx_to_cluster_array = sch.fcluster(linkage, cluster_distance_threshold, criterion='distance')
    idx = np.argsort(idx_to_cluster_array)

    if not inplace:
        corr_array = corr_array.copy()

    if isinstance(corr_array, pd.DataFrame):
        return corr_array.iloc[idx, :].T.iloc[idx, :], idx

    return corr_array[idx, :][:, idx], idx
'''

class similarity_matrix_tab(TabBase):

    def __init__(self, title='similarity'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def update_labels(self):
        labels = self.left_axis_cb.get_selected_result()
        if labels is not None:
            if self.current_label_resort is not None:
                labels = list(np.flip(np.array(labels)[self.current_label_resort]))
            labels = [(0.5 + i, l) for i, l in enumerate(labels)]
            self.plot.showAxis('left')
            ax = self.plot.getAxis('left')
            ax.setTickFont(QFont("Courier"))
            ax.setWidth(100)
            ax.setTicks([labels])

        labels = self.right_axis_cb.get_selected_result()
        if labels is not None:
            if self.current_label_resort is not None:
                labels = list(np.flip(np.array(labels)[self.current_label_resort]))
            labels = [(0.5 + i, l) for i, l in enumerate(labels)]
            self.plot.showAxis('right')
            ax = self.plot.getAxis('right')
            ax.setTickFont(QFont("Courier"))
            ax.setWidth(100)
            ax.setTicks([labels])


        #key = self.left_axis_cb.get_selected_key()
        #module = self.left_axis_cb.get_selected_module()
        #if module is not None:


        #for
        #r = self.current_group['Neuron_Reaction_Analysis', 0]
        #if r is not None:
        #    labels = r.get_representations()

        #    if self.current_label_resort is not None:
        #        labels = list(np.array(labels)[self.current_label_resort])



    def update_plot(self):
        key = self.content_cb.get_selected_key()
        module = self.content_cb.get_selected_module()
        if module is not None:
            image, self.current_label_resort = module.get_cluster_matrix(key)
            if image is not None:
                self.image_item.setImage(np.fliplr(image))#np.fliplr(image)
            #else:
            #    print('image is none')
        #else:
        #    print('module is none')


    def on_selected_neuron_changed(self, Network_UI):
        self.left_axis_cb.change_main_object(Network_UI.selected_neuron_group())
        self.content_cb.change_main_object(Network_UI.selected_neuron_group())
        self.right_axis_cb.change_main_object(Network_UI.selected_neuron_group())
        self.update_plot()

    def initialize(self, Network_UI):
        self.similarity_matrix_tab = Network_UI.Next_Tab(self.title)

        self.image_item, self.plot = Network_UI.Add_Image_Item(True, False, title='', tooltip_message='afferent synapse weights of selected neuron')

        #colors = [
        #    (0, 0, 0),
        #    (4, 5, 61),
        #    (84, 42, 55),
        #    (15, 87, 60),
        #    (208, 17, 141),
        #    (255, 255, 255)
        #]
        # color map
        #cmap = pg.ColorMap(pos=np.linspace(0.0, 1.0, 6), color=colors)
        #cmap = pg.get_cmap('CET-L9')
        cmap = generatePgColormap('viridis')

        self.image_item.setLookupTable(cmap.getLookupTable())
        #self.plot.setColorMap(cmap)

        #img = pg.ImageItem(image=noisy_data)  # create monochrome image from demonstration data
        #plot.addItem(img)  # add to PlotItem 'plot'
        #cm = pg.colormap.get('CET-L9')  # prepare a linear color map
        #bar = pg.ColorBarItem(values=(0, 20_000), cmap=cmap)  # prepare interactive color bar
        # Have ColorBarItem control colors of img and appear in 'plot':
        #bar.setImageItem(self.image_item, insert_in=self.plot)

        Network_UI.Next_H_Block()

        self.current_label_resort = None

        #self.similarity_images = {}
        #self.similarity_idxs = {}

        self.left_axis_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'labeler', first_entry=''))
        self.content_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'cluster_matrix_classifier'))
        self.right_axis_cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(Network_UI.network.NeuronGroups[0], 'labeler', first_entry=''))


        def update_btn_clicked(event):
            self.update_plot()
            self.update_labels()

        self.update_btn = Network_UI.Add_element(QPushButton('update'))
        self.update_btn.clicked.connect(update_btn_clicked)



    def update(self, Network_UI):
        if self.similarity_matrix_tab.isVisible():
            self.current_group = Network_UI.selected_neuron_group()



